"""
Device handling.
"""


from __future__ import annotations
from typing import Optional
from returnn.tensor import Tensor


def copy_to_device(x: Tensor, device: Optional[str]) -> Tensor:
    """
    Copy tensor to device.

    :param x: tensor
    :param device:
    :return: tensor on device
    """
    if not device:
        return x
    if x.raw_tensor is None:
        return x
    if x.device == device:
        return x
    # noinspection PyProtectedMember
    return x._raw_backend.copy_to_device(x, device)



"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
from typing import Tuple
import _setup_test_env  # noqa
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim
from rf_utils import run_model


def test_while_loop():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            def _cond(s: Tuple[Tensor, Tensor]):
                t, s_ = s
                if t.raw_tensor.__class__.__module__.startswith("torch"):
                    print("**", t.raw_tensor, rf.reduce_sum(s_, axis=s_.dims).raw_tensor)
                return rf.logical_and(rf.reduce_sum(s_, axis=s_.dims) < 50, t < time_dim.get_dim_value_tensor())

            def _body(s):
                t, s_ = s
                return t + 1, s_ + rf.abs(rf.gather(x, indices=t, axis=time_dim))

            _, final_s = rf.while_loop(
                _cond,
                _body,
                initial=(rf.zeros((), dtype=rf.get_default_array_index_dtype()), rf.zeros((batch_dim, in_dim))),
            )
            return final_s

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step, test_tensorflow=False)



class BundleFile(object):
    """Holds paths to HDF dataset files."""

    def __init__(self, filePath):
        """Reads paths to HDF dataset files from a bundle file.
        Example of contents of a bundle file:

        /work/asr2/ryndin/crnnRegressionSpeechEnhancemenent/data/data_tr05_real_1_100.hdf
        /work/asr2/ryndin/crnnRegressionSpeechEnhancemenent/data/data_tr05_real_2_100.hdf
        /work/asr2/ryndin/crnnRegressionSpeechEnhancemenent/data/data_tr05_real_3_100.hdf
        /work/asr2/ryndin/crnnRegressionSpeechEnhancemenent/data/data_tr05_real_4_100.hdf
        /work/asr2/ryndin/crnnRegressionSpeechEnhancemenent/data/data_tr05_real_5_100.hdf
        /work/asr2/ryndin/crnnRegressionSpeechEnhancemenent/data/data_tr05_real_6_100.hdf
        /work/asr2/ryndin/crnnRegressionSpeechEnhancemenent/data/data_tr05_simu_1_100.hdf
        /work/asr2/ryndin/crnnRegressionSpeechEnhancemenent/data/data_tr05_simu_2_100.hdf
        /work/asr2/ryndin/crnnRegressionSpeechEnhancemenent/data/data_tr05_simu_3_100.hdf
        /work/asr2/ryndin/crnnRegressionSpeechEnhancemenent/data/data_tr05_simu_4_100.hdf
        /work/asr2/ryndin/crnnRegressionSpeechEnhancemenent/data/data_tr05_simu_5_100.hdf
        /work/asr2/ryndin/crnnRegressionSpeechEnhancemenent/data/data_tr05_simu_6_100.hdf

        :type filePath: str
        :param filePath: path to a bundle file which contains paths to HDF
                         dataset files. One path per line.
        """
        self._filePath = filePath
        self._datasetFilesPaths = []
        self._readDatasetFilesPaths()

    def _readDatasetFilesPaths(self):
        """Reads paths to HDF dataset files from a bundle file."""
        with open(self._filePath, "r") as bundleFile:
            self._datasetFilesPaths = filter(
                lambda f: bool(f),  # filter off empty lines
                map(lambda l: l.strip(), bundleFile.readlines()),  # strip spaces from left and right
            )

    @property
    def datasetFilePaths(self):
        """Paths to HDF dataset files.

        :rtype: list of str
        :return: Paths to HDF dataset files.
        """
        return self._datasetFilesPaths

    @property
    def numberOfDatasetFiles(self):
        """Number of HDF dataset files.

        :rtype: int
        :return: Number of HDF dataset files.
        """
        return len(self._datasetFilesPaths)



"""
Converts a module from a config to ONNX. For that, it uses get_model() which must be available in the config
and creates dummy data to be forwarded to the model.

Since get_model() can return either a torch.nn.Module or a rf.Module, both cases must be taken into account.
"""

from __future__ import annotations
import torch
from typing import Callable, Optional, Dict
import argparse
import os
import numpy

from returnn.config import Config
from returnn.log import log
from returnn.tensor import Dim, Tensor, TensorDict

# noinspection PyProtectedMember
from returnn.torch.frontend.bridge import _RFModuleAsPTModule
import returnn.frontend as rf
import returnn.util.basic as util
from returnn.torch.data.tensor_utils import tensor_numpy_to_torch_
import returnn.__main__ as rnn


config = None  # type: Optional[Config]


def init(config_filename: str, checkpoint: str, log_verbosity: int, device: str):
    """
    :param config_filename: Filename to config file.
    :param checkpoint: Filename to the trained model.
    :param log_verbosity: 5 for all seqs (default: 4)
    :param device:
    """
    assert os.path.exists(checkpoint), "The specified checkpoint doesn't exist."
    rnn.init_better_exchook()
    rnn.init_thread_join_hack()
    assert os.path.exists(config_filename), "The specified config doesn't exist."
    print("Using config file %r." % config_filename)
    rnn.init_config(
        config_filename=config_filename,
        extra_updates={
            "log": None,
            "log_verbosity": log_verbosity,
            "task": __file__,  # just extra info for the config
            "device": device,
        },
    )
    global config
    config = rnn.config
    rnn.init_log()
    print("RETURNN frontend module to ONNX conversion.", file=log.v1)
    rnn.returnn_greeting()
    config.typed_dict.setdefault("backend", "torch")
    rnn.init_backend_engine()
    assert util.BackendEngine.is_torch_selected(), "For now only the torch backend is supported."
    rnn.init_faulthandler()


class ForwardModulePT(torch.nn.Module):
    """
    Wrapper of a PyTorch module that's meant to call forward_step from the config when called.
    """

    def __init__(self, pt_module: torch.nn.Module, forward_step: Callable, extern_data: TensorDict):
        """
        :param pt_module: RF module as obtained from the config.
        :param forward_step: forward_step function as obtained from the config.
        :param extern_data:
        """
        super().__init__()

        self.model = pt_module
        self.forward_step_func = forward_step
        self.extern_data = extern_data

    def __call__(self, data: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Wrapper to forward_step from the config.
        """
        extern_data = self.extern_data.copy_template()
        extern_data.assign_from_raw_tensor_dict_(data)
        self.forward_step_func(model=self.model, extern_data=extern_data)
        return rf.get_run_ctx().outputs.as_raw_tensor_dict()


class ForwardModuleRF(_RFModuleAsPTModule):
    """
    Wrapper of a RETURNN frontend module that's meant to call forward_step from the config when called.
    """

    def __init__(self, rf_module: rf.Module, forward_step: Callable, extern_data: TensorDict):
        """
        :param rf_module: RF module as obtained from the config.
        :param forward_step: forward_step function as obtained from the config.
        :param extern_data:
        """
        super().__init__(rf_module)

        self.forward_step_func = forward_step
        self.extern_data = extern_data

    def __call__(self, data: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Wrapper to forward_step from the config.
        """
        extern_data = self.extern_data.copy_template()
        extern_data.assign_from_raw_tensor_dict_(data)
        self.forward_step_func(model=self.rf_module, extern_data=extern_data)
        return rf.get_run_ctx().outputs.as_raw_tensor_dict()


def main():
    """
    Main entry point
    """
    parser = argparse.ArgumentParser(description="Converts a RF/PT module to ONNX.")
    parser.add_argument(
        "config", type=str, help="Filename to config file. Must have get_model(). Can optionally have export()."
    )
    parser.add_argument("checkpoint", type=str, help="Checkpoint to RF module, considering the backend.")
    parser.add_argument("out_onnx_filename", type=str, help="Filename of the final ONNX model.")
    parser.add_argument("--verbosity", default=4, type=int, help="5 for all seqs (default: 4)")
    parser.add_argument("--device", type=str, default="cpu", help="'cpu' (default) or 'gpu'.")
    args = parser.parse_args()

    init(config_filename=args.config, checkpoint=args.checkpoint, log_verbosity=args.verbosity, device=args.device)
    rf.init_forward_step_run_ctx()
    rf.set_random_seed(42)

    get_model_func = config.typed_value("get_model")
    assert get_model_func, "get_model() isn't specified in the config passed as a parameter."
    model = get_model_func()
    loaded_checkpoint = torch.load(args.checkpoint)

    is_rf_module = isinstance(model, rf.Module)
    is_pt_module = isinstance(model, torch.nn.Module)
    assert (
        is_rf_module or is_pt_module
    ), "The module returned by get_model() isn't a returnn.frontend.Module or a torch.nn.Module."

    export_func = config.typed_value("export") or torch.onnx.export
    forward_step_func = config.typed_value("forward_step")
    assert forward_step_func is not None, "forward_step() must be defined in the config."

    extern_data_dict = config.typed_value("extern_data")
    extern_data = TensorDict()
    extern_data.update(extern_data_dict, auto_convert=True)
    extern_data.reset_content()

    rnd = numpy.random.RandomState(42)
    for v in extern_data.data.values():
        _fill_random(v, rnd=rnd)
    for v in extern_data.data.values():
        tensor_numpy_to_torch_(v)
    extern_data_raw = extern_data.as_raw_tensor_dict()

    if is_pt_module:
        model.load_state_dict(loaded_checkpoint["model"])
        model.eval()
        pt_model_fwd = ForwardModulePT(model, forward_step_func, extern_data)
    else:
        pt_model_fwd = ForwardModuleRF(model, forward_step_func, extern_data)
        pt_model_fwd.load_state_dict(loaded_checkpoint["model"])
        pt_model_fwd.eval()

    model_outputs_dict = config.typed_value("model_outputs")
    model_outputs = TensorDict()
    model_outputs.update(model_outputs_dict, auto_convert=True)
    model_outputs_raw_keys = []
    for k, v in model_outputs.data.items():
        model_outputs_raw_keys.append(k)
        for i, dim in enumerate(v.dims):
            if dim.is_batch_dim() or dim.is_dynamic():
                model_outputs_raw_keys.append(f"{k}:size{i}")

    dynamic_axes = {}
    for k, v in list(extern_data.data.items()) + list(model_outputs.data.items()):
        dynamic_axes[k] = {i: dim.name for i, dim in enumerate(v.dims) if dim.is_dynamic() or dim.is_batch_dim()}
        for i, dim in enumerate(v.dims):
            if dim.dyn_size_ext:
                dynamic_axes[f"{k}:size{i}"] = {
                    j: dim_.name
                    for j, dim_ in enumerate(dim.dyn_size_ext.dims)
                    if dim_.is_dynamic() or dim_.is_batch_dim()
                }

    export_func(
        pt_model_fwd,
        (extern_data_raw, {}),
        f=args.out_onnx_filename,
        verbose=True,
        input_names=list(extern_data_raw.keys()),
        output_names=model_outputs_raw_keys,
        dynamic_axes=dynamic_axes,
    )


def _fill_random(
    x: Tensor,
    *,
    min_val: int = 0,
    max_val: Optional[int] = None,
    rnd: numpy.random.RandomState,
    dyn_dim_max_sizes: Optional[Dict[Dim, int]] = None,
) -> bool:
    """fill. return whether sth was filled"""
    if dyn_dim_max_sizes is None:
        dyn_dim_max_sizes = {}
    filled = False
    while True:
        have_unfilled = False
        filled_this_round = False

        for dim in x.dims:
            if dim.is_batch_dim() and not dim.dyn_size_ext:
                dim.dyn_size_ext = Tensor("batch", [], dtype="int32")
            if not dim.dyn_size_ext:
                continue
            if _fill_random(
                dim.dyn_size_ext,
                min_val=2,
                max_val=dyn_dim_max_sizes.get(dim, None),
                rnd=rnd,
                dyn_dim_max_sizes=dyn_dim_max_sizes,
            ):
                if dim in dyn_dim_max_sizes:
                    # Make sure at least one of the dyn sizes matches the max size.
                    i = rnd.randint(0, dim.dyn_size_ext.raw_tensor.size)
                    dim.dyn_size_ext.raw_tensor.flat[i] = dyn_dim_max_sizes[dim]
                filled = True
                filled_this_round = True
            if dim.dyn_size_ext.raw_tensor is None:
                have_unfilled = True
            elif not isinstance(dim.dyn_size_ext.raw_tensor, numpy.ndarray):
                have_unfilled = True

        if have_unfilled:
            assert filled_this_round, f"should have filled something, {x}"

        if not have_unfilled:
            break

    if x.raw_tensor is not None:
        if not isinstance(x.raw_tensor, numpy.ndarray):
            x.raw_tensor = None

    if x.raw_tensor is None:
        shape = [d.get_dim_value() for d in x.dims]
        if x.dtype.startswith("int"):
            if max_val is None:
                max_val = rnd.randint(5, 20)
            if x.sparse_dim and x.sparse_dim.dimension is not None:
                max_val = x.sparse_dim.dimension
            x.raw_tensor = rnd.randint(min_val, max_val, size=shape, dtype=x.dtype)
        elif x.dtype.startswith("float"):
            x.raw_tensor = rnd.normal(0.0, 1.0, size=shape).astype(x.dtype)
        elif x.dtype.startswith("complex"):
            real = rnd.normal(0.0, 1.0, size=shape)
            imag = rnd.normal(0.0, 1.0, size=shape)
            x.raw_tensor = (real + 1j * imag).astype(x.dtype)
        else:
            raise NotImplementedError(f"not implemented for {x} dtype {x.dtype}")
        filled = True

    assert isinstance(x.raw_tensor, numpy.ndarray)

    return filled


if __name__ == "__main__":
    main()



"""
stft etc
"""


from __future__ import annotations
from typing import Optional, Union, Tuple
import numpy
import functools
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf


def stft(
    x: Tensor,
    *,
    in_spatial_dim: Dim,
    frame_step: int,
    frame_length: int,
    fft_length: Optional[int] = None,
    window_use_frame_length: bool = True,
    align_window_left: bool = True,
    window_enforce_even: bool = True,
    out_spatial_dim: Optional[Dim] = None,
    out_dim: Optional[Dim] = None,
) -> Tuple[Tensor, Dim, Dim]:
    """
    Calculate the short-time Fourier transform (STFT) of a signal.

    We currently always use the Hann window.

    Note that there are inconsistencies between frameworks (e.g. PyTorch vs TensorFlow).
    We have options where you can explicitly specify the behavior,
    to match either PyTorch or TensorFlow.
    See here for some discussion and demonstration:
    https://github.com/pytorch/pytorch/issues/100177
    https://github.com/librosa/librosa/issues/1701
    https://github.com/albertz/playground/blob/master/tf_pt_stft.py

    :param x: (..., in_spatial_dim, ...). any other dims are treated as batch dims.
    :param in_spatial_dim:
    :param frame_length: win_length in PT/librosa, frame_size in StftLayer, frame_length in TF
    :param frame_step: hop_length in PT/librosa, frame_shift in StftLayer, frame_step in TF
    :param fft_length: n_fft in PT/librosa, fft_size in StftLayer, fft_length in TF.
        If not given, will use frame_length.
    :param window_use_frame_length: Whether to use window size = frame_length.
        If False, uses window size = fft_length.
        This has only an effect if frame_length != fft_length.
        This has an effect on the output seq length.
        If True: out_seq_len = ⌈(T - frame_length + 1) / frame_step⌉ (T = in_seq_len).
        If False: out_seq_len = ⌈(T - fft_length + 1) / frame_step⌉.
        Note that in TF/SciPy, the behavior matches to window_use_frame_length=True,
        but in PT/librosa, the behavior matches to window_use_frame_length=False.
    :param align_window_left: whether to align the window to the left inside the potential wider fft_length window.
        This only has an effect if frame_length != fft_length.
        If True, the window will be aligned to the left and the right side will be padded with zeros.
        Then effectively the remaining phases are not used.
        If False, the window will be aligned to the center and the left and right side will be padded with zeros.
        Then effectively the beginning and end phases are not used.
        In TF/SciPy, the behavior matches to align_window_left=True,
        but in PT/librosa, the behavior matches to align_window_left=False.
    :param window_enforce_even: enforce that the window size is even, potentially lowering the window size by 1.
        This only has an effect if the window size is uneven.
        In TF, the behavior matches to window_enforce_even=True,
        but in most other frameworks, the behavior matches to window_enforce_even=False.
    :param out_spatial_dim:
    :param out_dim:
    :return: (stft, out_spatial_dim, out_dim)
    """
    fft_length = fft_length or frame_length
    if out_dim is None:
        out_dim = Dim(fft_length // 2 + 1, name="stft-freq")
    if out_spatial_dim is None:
        from .conv import make_conv_out_spatial_dims

        (out_spatial_dim,) = make_conv_out_spatial_dims(
            [in_spatial_dim], filter_size=frame_length, strides=frame_step, padding="valid"
        )
    # noinspection PyProtectedMember
    return (
        x._raw_backend.stft(
            x,
            in_spatial_dim=in_spatial_dim,
            frame_step=frame_step,
            frame_length=frame_length,
            fft_length=fft_length,
            window_use_frame_length=window_use_frame_length,
            align_window_left=align_window_left,
            window_enforce_even=window_enforce_even,
            out_spatial_dim=out_spatial_dim,
            out_dim=out_dim,
        ),
        out_spatial_dim,
        out_dim,
    )


def mel_filterbank(
    x: Tensor,
    *,
    in_dim: Dim,
    out_dim: Dim,
    sampling_rate: Union[int, float],
    fft_length: Optional[int] = None,
    f_min: Optional[Union[int, float]] = None,
    f_max: Optional[Union[int, float]] = None,
):
    """
    Applies the Mel filterbank to the input.

    :param x:
    :param in_dim: expected to be fft_length // 2 + 1. E.g. via :func:`stft`.
    :param out_dim: nr of mel filters.
    :param sampling_rate:
    :param fft_length: fft_size, n_fft. Should match fft_length from :func:`stft`.
        If not given, infer this from in_dim, as (in_dim - 1) * 2.
    :param f_min:
    :param f_max:
    :return:
    """
    if not fft_length:
        fft_length = (in_dim.dimension - 1) * 2
    f_min = f_min or 0
    f_max = f_max or sampling_rate / 2.0
    # noinspection PyProtectedMember
    backend = x._raw_backend
    cache_key = (f_min, f_max, sampling_rate, fft_length, out_dim.dimension, backend, x.dtype, x.device)
    if cache_key in _mel_filter_bank_matrix_cache:
        filter_bank_matrix = _mel_filter_bank_matrix_cache[cache_key]
    else:
        filter_bank_matrix_np = _mel_filter_bank_matrix_np(
            f_min=f_min, f_max=f_max, sampling_rate=sampling_rate, fft_size=fft_length, nr_of_filters=out_dim.dimension
        )
        filter_bank_matrix_np = filter_bank_matrix_np.astype(x.dtype)
        filter_bank_matrix = rf.convert_to_tensor(filter_bank_matrix_np, dims=(in_dim, out_dim), _backend=backend)
        filter_bank_matrix = rf.copy_to_device(filter_bank_matrix, x.device)
        if backend.executing_eagerly():
            if len(_mel_filter_bank_matrix_cache) > 100:
                # Very simple cache management. No LRU logic or anything like that.
                _mel_filter_bank_matrix_cache.clear()
            _mel_filter_bank_matrix_cache[cache_key] = filter_bank_matrix
    return rf.matmul(x, filter_bank_matrix, reduce=in_dim)


# Used for eager frameworks. See mel_filterbank.
_mel_filter_bank_matrix_cache = {}


@functools.lru_cache()
def _mel_filter_bank_matrix_np(
    *,
    f_min: Union[int, float],
    f_max: Union[int, float],
    sampling_rate: Union[int, float],
    fft_size: int,
    nr_of_filters: int,
) -> numpy.ndarray:
    """
    Returns the filter matrix which yields the mel filter bank features, when applied to the spectrum as
    matmul(freqDom, filterMatrix), where freqDom has dimension (time, frequency)
    and filterMatrix is the matrix returned
    by this function.
    The filter matrix is computed according to equation 6.141 in
    [Huang & Acero+, 2001] "Spoken Language Processing - A Guide to Theroy, Algorithm, and System Development"

    :param float|int f_min: minimum frequency
    :param float|int f_max: maximum frequency
    :param float sampling_rate: sampling rate of audio signal
    :param int fft_size: dimension of discrete fourier transformation
    :param int nr_of_filters: number of mel frequency filter banks to be created
    :return: shape=(fft_size // 2 + 1, nr_of_filters), matrix yielding the mel frequency cepstral coefficients
    """

    def mel_scale(freq):
        """
        returns the respective value on the mel scale

        :param float freq: frequency value to transform onto mel scale
        :rtype: float
        """
        return 1125.0 * numpy.log(1 + float(freq) / 700)

    def inv_mel_scale(mel_val):
        """
        returns the respective value in the frequency domain

        :param float mel_val: value in mel domain
        :rtype: float
        """
        return 700.0 * (numpy.exp(float(mel_val) / 1125) - 1)

    # noinspection PyShadowingNames
    def filter_center(filter_id, f_min, f_max, sampling_rate, fft_size, nr_of_filters):
        """
        :param int filter_id: filter to compute the center frequency for
        :param float|int f_min: minimum frequency
        :param float|int f_max: maximum frequency
        :param float|int sampling_rate: sampling rate of audio signal
        :param int fft_size: dimension of discrete fourier transformation
        :param int nr_of_filters: number of mel frequency filter banks to be created
        :rtype: float
        :return: center frequency of filter
        """
        return (float(fft_size) / sampling_rate) * inv_mel_scale(
            mel_scale(f_min) + filter_id * ((mel_scale(f_max) - mel_scale(f_min)) / (nr_of_filters + 1))
        )

    filt_cent = numpy.zeros(shape=(nr_of_filters + 2,), dtype=numpy.float32)
    for i1 in range(nr_of_filters + 2):
        filt_cent[i1] = filter_center(i1, f_min, f_max, sampling_rate, fft_size, nr_of_filters)
    f_mat = numpy.zeros(shape=(fft_size // 2 + 1, nr_of_filters))
    for i1 in range(f_mat.shape[0]):
        for i2 in range(1, nr_of_filters + 1):
            if (i1 > filt_cent[i2 - 1]) and (i1 < filt_cent[i2 + 1]):
                if i1 < filt_cent[i2]:
                    num = i1 - filt_cent[i2 - 1]
                    denom = filt_cent[i2] - filt_cent[i2 - 1]
                else:
                    num = filt_cent[i2 + 1] - i1
                    denom = filt_cent[i2 + 1] - filt_cent[i2]
                el_val = num / denom
            else:
                el_val = 0
            f_mat[i1, i2 - 1] = el_val

    return f_mat



"""
Provides the main class for logging, :class:`Log`, and some helpers.
"""

from __future__ import annotations

import logging
import os
import sys
import io
import threading
from threading import RLock
import contextlib
import string
import typing

PY3 = sys.version_info[0] >= 3


class Stream:
    """
    Simple stream wrapper, which provides :func:`write` and :func:`flush`.
    """

    # noinspection PyShadowingNames
    def __init__(self, log, lvl):
        """
        :type log: logging.Logger
        :type lvl: int
        """
        self.buf = io.StringIO()
        self.log = log
        self.lvl = lvl
        self.lock = RLock()

    def write(self, msg):
        """
        :param str msg:
        """
        with self.lock:
            if msg == "\n":
                self.flush()
            else:
                self.buf.write(msg)

    def flush(self):
        """
        Flush, i.e. writes to the log.
        """
        with self.lock:
            self.buf.flush()
            self.log.log(self.lvl, self.buf.getvalue())
            self.buf.truncate(0)
            # truncate does not change the current position.
            # In Python 2.7, it incorrectly does. See: https://bugs.python.org/issue30250
            self.buf.seek(0)


class Log:
    """
    The main logging class.
    """

    def __init__(self):
        self.initialized = False
        self.filename = None  # type: typing.Optional[str]
        self.v = None  # type: typing.Optional[typing.List[logging.Logger]]
        self.verbose = None  # type: typing.Optional[typing.List[bool]]
        self.v1 = None  # type: typing.Optional[Stream]
        self.v2 = None  # type: typing.Optional[Stream]
        self.v3 = None  # type: typing.Optional[Stream]
        self.v4 = None  # type: typing.Optional[Stream]
        self.v5 = None  # type: typing.Optional[Stream]
        self._printed_warning_history = set()  # type: typing.Set[str]

    def initialize(self, logs=None, verbosity=None, formatter=None, propagate=False):
        """
        This resets and configures the "returnn" logger.

        :param list[str|logging.Handler] logs: "stdout", "|<pipe-cmd>", "<filename>"|"<filename>$date<ext>".
          "stdout" is always added when propagate=False.
        :param list[int] verbosity: levels 0-5 for the log handlers
        :param list[str] formatter: 'default', 'timed', 'raw' or 'verbose', for the log handlers
        :param bool propagate:
        """
        if formatter is None:
            formatter = []
        if verbosity is None:
            verbosity = []
        if logs is None:
            logs = []
        self.initialized = True
        fmt = {
            "default": logging.Formatter("%(message)s"),
            "timed": logging.Formatter("%(asctime)s %(message)s", datefmt="%Y-%m-%d,%H:%M:%S.%MS"),
            "raw": logging.Formatter("%(message)s"),
            "verbose": logging.Formatter("%(levelname)s - %(asctime)s %(message)s", datefmt="%Y-%m-%d,%H:%M:%S.%MS"),
        }
        logger = logging.getLogger("returnn")
        # Note on propagation:
        # This is not so clear. By default, the root logger anyway does nothing.
        # However, if you mix RETURNN with other code, which might setup the root logger
        # (e.g. via logging.basicConfig(...)), then there is some root logger,
        # and maybe we should also use it.
        # But at this point here, we might not know about this
        # -- maybe the user would call logging.basicConfig(...) at some later point.
        # In any case, if there is a root logger and we would propagate,
        # we should not add "stdout" here,
        # although that might depend on the root logger level and handlers.
        # For now, the default is to just disable propagation, to keep that separated
        # and avoid any such problems.
        logger.propagate = propagate
        # Reset handler list, in case we have initialized some earlier (e.g. multiple log.initialize() calls).
        logger.handlers = []
        self.v = [logger] * 6  # no need for separate loggers, we do all via log levels
        if "stdout" not in logs and not propagate:
            logs.append("stdout")
        if len(formatter) == 1:
            # if only one format provided, use it for all logs
            formatter = [formatter[0]] * len(logs)
        # Define own level names. In reverse order, such that the name by default still has the default behavior.
        logging.addLevelName(logging.DEBUG + 2, "DEBUG")
        logging.addLevelName(logging.DEBUG + 1, "DEBUG")
        logging.addLevelName(logging.DEBUG + 0, "DEBUG")
        logging.addLevelName(logging.INFO + 1, "INFO")
        logging.addLevelName(logging.INFO + 0, "INFO")
        _VerbosityToLogLevel = {
            0: logging.ERROR,
            1: logging.INFO + 1,
            2: logging.INFO,
            3: logging.DEBUG + 2,
            4: logging.DEBUG + 1,
            5: logging.DEBUG,
        }
        self.verbose = [False] * 6
        for i in range(len(logs)):
            t = logs[i]
            v = 3
            if i < len(verbosity):
                v = verbosity[i]
            elif len(verbosity) == 1:
                v = verbosity[0]
            assert v <= 5, "invalid verbosity: " + str(v)
            for j in range(v + 1):
                self.verbose[j] = True
            f = fmt["default"] if i >= len(formatter) or formatter[i] not in fmt else fmt[formatter[i]]
            if isinstance(t, logging.Handler):
                handler = t
            elif t == "stdout":
                handler = StdoutHandler()
            elif t.startswith("|"):  # pipe-format
                proc_cmd = t[1:].strip()
                from subprocess import Popen, PIPE

                proc = Popen(proc_cmd, shell=True, stdin=PIPE)
                handler = logging.StreamHandler(proc.stdin)
            elif os.path.isdir(os.path.dirname(t)):
                if "$" in t:
                    from returnn.util.basic import get_utc_start_time_filename_part

                    t = string.Template(t).substitute(date=get_utc_start_time_filename_part())
                self.filename = t
                handler = logging.FileHandler(t)
            else:
                assert False, "invalid log target %r" % t
            handler.setLevel(_VerbosityToLogLevel[v])
            handler.setFormatter(f)
            logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)
        if not logger.handlers:
            logger.addHandler(logging.NullHandler())
        self.v1 = Stream(self.v[1], _VerbosityToLogLevel[1])
        self.v2 = Stream(self.v[2], _VerbosityToLogLevel[2])
        self.v3 = Stream(self.v[3], _VerbosityToLogLevel[3])
        self.v4 = Stream(self.v[4], _VerbosityToLogLevel[4])
        self.v5 = Stream(self.v[5], _VerbosityToLogLevel[5])

    def init_by_config(self, config):
        """
        :param returnn.config.Config config:
        """
        logs = config.list("log", [])
        log_verbosity = config.int_list("log_verbosity", [])
        log_format = config.list("log_format", [])
        if config.is_true("use_horovod"):
            import returnn.tf.horovod

            hvd = returnn.tf.horovod.get_ctx(config=config)
            new_logs = []
            for fn in logs:
                fn_prefix, fn_ext = os.path.splitext(fn)
                fn_ext = ".horovod-%i-%i%s" % (hvd.rank(), hvd.size(), fn_ext)
                new_logs.append(fn_prefix + fn_ext)
            logs = new_logs
        self.initialize(logs=logs, verbosity=log_verbosity, formatter=log_format)

    def print_warning(self, text, prefix_text="WARNING:", extra_text=None):
        """
        Write a warning to log.v2. Does not write repeated warnings.

        :param str text:
        :param str prefix_text:
        :param str|None extra_text:
        """
        if text in self._printed_warning_history:
            return
        self._printed_warning_history.add(text)
        print(prefix_text, text, file=log.v2)
        if extra_text:
            print(extra_text, file=log.v2)

    def print_deprecation_warning(self, text, behavior_version=None):
        """
        Write a deprecation warning to log.v2. Does not write repeated warnings.

        :param str text:
        :param int|None behavior_version: if this deprecation is already covered by a behavior_version check
        """
        if behavior_version:
            behavior_text = "This will be disallowed with behavior_version %d." % behavior_version
        else:
            behavior_text = "This might be disallowed with a future behavior_version."
        self.print_warning(text, prefix_text="DEPRECATION WARNING:", extra_text=behavior_text)

    def flush(self):
        """
        Flush all streams.
        """
        for stream in [self.v1, self.v2, self.v3, self.v4, self.v5]:
            if stream:
                stream.flush()


log = Log()


class StdoutHandler(logging.StreamHandler):
    """
    This class is like a StreamHandler using sys.stdout, but always uses
    whatever sys.stdout is currently set to rather than the value of
    sys.stdout at handler construction time.

    Copied and adopted from logging._StderrHandler.
    """

    @property
    def stream(self):
        """
        stream
        """
        return sys.stdout

    @stream.setter
    def stream(self, stream):
        pass  # ignore


class StreamThreadLocal(threading.local):
    """
    This will just buffer everything, thread-locally, and not forward it to any stream.
    The idea is that multiple tasks will run in multiple threads and you want to catch all the logging/stdout
    of each to not clutter the output, and also you want to keep it separate for each.
    """

    def __init__(self):
        self.buf = io.StringIO()

    def write(self, msg):
        """
        :param str msg:
        """
        self.buf.write(msg)

    def flush(self):
        """
        Ignored.
        """


class StreamDummy:
    """
    This will just discard any data.
    """

    def write(self, msg):
        """
        Ignored.

        :param str msg:
        """
        pass

    def flush(self):
        """
        Ignored.
        """


@contextlib.contextmanager
def wrap_log_streams(alternative_stream, also_sys_stdout=False, tf_log_verbosity=None):
    """
    :param StreamThreadLocal|StreamDummy alternative_stream:
    :param bool also_sys_stdout: wrap sys.stdout as well
    :param int|str|None tf_log_verbosity: e.g. "WARNING"
    :return: context manager which yields (original info stream v1, alternative_stream)
    """
    v_attrib_keys = ["v%i" % i for i in range(6)] + ["error"]
    # Store original values.
    orig_v_list = log.v
    orig_v_attribs = {key: getattr(log, key) for key in v_attrib_keys}
    orig_stdout = sys.stdout
    log.v = [alternative_stream] * len(orig_v_list)
    for key in v_attrib_keys:
        setattr(log, key, alternative_stream)
    if also_sys_stdout:
        sys.stdout = alternative_stream
    orig_tf_log_verbosity = None
    if tf_log_verbosity is not None:
        import returnn.tf.compat as tf_compat

        orig_tf_log_verbosity = tf_compat.v1.logging.get_verbosity()
        tf_compat.v1.logging.set_verbosity(tf_log_verbosity)
    try:
        yield orig_v_attribs["v1"], alternative_stream
    finally:
        # Restore original values.
        log.v = orig_v_list
        for key, value in orig_v_attribs.items():
            setattr(log, key, value)
        if also_sys_stdout:
            sys.stdout = orig_stdout
        if tf_log_verbosity is not None:
            import returnn.tf.compat as tf_compat

            tf_compat.v1.logging.set_verbosity(orig_tf_log_verbosity)
